/* BEGIN software license
 *
 * msXpertSuite - mass spectrometry software suite
 * -----------------------------------------------
 * Copyright(C) 2009,...,2021 Filippo Rusconi
 *
 * http://www.msxpertsuite.org
 *
 * This file is part of the msXpertSuite project.
 *
 * The msXpertSuite project is the successor of the massXpert project. This
 * project now includes various independent modules:
 *
 * - massXpert, model polymer chemistries and simulate mass spectrometric data;
 * - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * END software license
 */


#pragma once


#include <QObject>

#include "../../trace/trace.h"
#include "../../exportinmportconfig.h"
#include "filternameinterface.h"


namespace pappso
{


class FilterLowIntensitySignalRemoval;

typedef std::shared_ptr<FilterLowIntensitySignalRemoval>
  FilterLowIntensitySignalRemovalSPtr;
typedef std::shared_ptr<const FilterLowIntensitySignalRemoval>
  FilterLowIntensitySignalRemovalCstSPtr;


/**
 * @brief Redefines the floor intensity of the Trace
 *
 * The amplitude of the trace is computed (maxValue - minValue)
 * Its fraction is calculated = amplitude * (percentage / 100)
 * The threshold value is computed as (minValue + fraction)
 *
 * When the values to be filtered are below that threshold they acquire that
 * threshold value.
 *
 * When the values to be filtered are above that threshold they remain
 * unchanged.
 *
 * This effectively re-floors the values to threshold.
 */
class PMSPP_LIB_DECL FilterLowIntensitySignalRemoval
  : public FilterNameInterface
{
  // We need this std::pair entity:
  using IndexIntensity = std::pair<std::size_t, double>;

  public:
  FilterLowIntensitySignalRemoval(double mean,
                                  double std_dev,
                                  double threshold);
  FilterLowIntensitySignalRemoval(const QString &parameters);
  FilterLowIntensitySignalRemoval(const FilterLowIntensitySignalRemoval &other);

  virtual ~FilterLowIntensitySignalRemoval();

  FilterLowIntensitySignalRemoval &
  operator=(const FilterLowIntensitySignalRemoval &other);

  Trace &filter(Trace &data_points) const override;

  double getThreshold() const;
  QString name() const override;

  QString toString() const override;

  protected:
  void buildFilterFromString(const QString &strBuildParams) override;
  Trace &nonConstFilter(Trace &data_points);

  private:
  const double nan = std::numeric_limits<double>::quiet_NaN();

  double m_threshold;
  double m_noiseMean;
  double m_noiseStdDev;

  constexpr static double INTRA_CLUSTER_INTER_PEAK_DISTANCE = 1.1;

  Trace m_helperTraceCopy;
  const std::size_t m_minIntPointCount   = 5;
  const std::size_t m_minIntStdDevFactor = 2;

  double m_min;
  double m_max;
  double m_minMean;
  double m_minStdDev;
  double m_noiseLevel;

  bool m_seen_upward_phase = false;

  IndexIntensity m_prevApex = std::pair(0, nan);
  IndexIntensity m_curApex  = std::pair(0, nan);
  IndexIntensity m_cur     = std::pair(0, nan);
  IndexIntensity m_prev     = std::pair(0, nan);

  using ApexSPtr = std::shared_ptr<IndexIntensity>;
  using Cluster = std::vector<ApexSPtr>;
  using ClusterSPtr = std::shared_ptr<std::vector<ApexSPtr>>;

  // The isotopic clusters are stored as vectors of IndexIntensity shared_ptr.
  std::vector<ClusterSPtr> m_isotopicClusters;

  std::size_t detectIsotopicClusters(const Trace &trace);
};
} // namespace pappso
